/**********************************************************************
;       Copyright 2005 Summit Soft Consulting
;
;                       RESTRICTED RIGHTS LEGEND
;                       ------------------------
;
; Use, duplication, or disclosure by any commercial industry (public
; or private), private individual, or by any Government Agency without
; a express written consent of release from Summit Soft Consulting
; is subject to restrictions as set forth in paragraph (b) (3) (B) of
; the Rights in Technical Data and Computer Software clause in
; DAR 7-104.9 (a).  Manufacturer is: Summit Soft Consulting,
; 26895 Aliso Creek Rd. Suite B504, Aliso Viejo, CA 92656,
; http://www.SummitSoftConsulting.com
;**********************************************************************/

#include "StdAfx.h"
#include "BarPane.h"
#include "bcmenu.h"

BOOL CBarPane::m_bHexadecimal = TRUE;
HexEditMode CBarPane::m_HexEditMode = hemByte;

CBarPane::CBarPane(CPciFunctionView * pPciFunctionView, DWORD dwBarIndex)
{
	// Save the pointer to our parent.
	m_pPciFunctionView = pPciFunctionView;

	// Save the BAR index. We use the index to know
	// which controls to modify in the dialog resource.
	m_dwBarIndex = dwBarIndex;

	// We use this cached PCI function 
	// to be able to switch display mode.
	m_pCachedPciFunction = NULL;	

	// Make one-time initializations.
	m_InitializeBarListCtrl();
}

CBarPane::~CBarPane(void)
{
	// We need to 'disconnect' from the list box defined in the 
	// dialog template. We will connect to it again the next 
	// time a tab pane for the same tab is created.
	m_editListCtrl.UnsubclassWindow();
}

BOOL CBarPane::m_InitializeBarListCtrl()
{
	// Subclass the list control defined in the dialog resources.
	BOOL bSuccess = m_editListCtrl.SubclassDlgItem(IDC_BAR0_PANE_LIST_CTRL + m_dwBarIndex, m_pPciFunctionView);
	_ASSERTE(bSuccess);

	// Set the background color of the list control to the nice yellow...
	m_editListCtrl.SetBkColor(RGB(0xF7, 0xF3, 0xE9));

	// Configure various styles of the list control.
	m_editListCtrl.SetExtendedStyle(LVS_EX_GRIDLINES);

	return TRUE;
}

void CBarPane::m_ConfigureDisplayMode()
{
	// Disable the data callbacks while we modify the list control layout. 
	// If we don't do this the display will flicker terribly as data is 
	// added while we are in the loop below.
	m_editListCtrl.SetItemCount(0);

	// Remove the previous columns, if any.
	while(m_editListCtrl.DeleteColumn(0))
		;

	// Add the address column to the list control.
	m_editListCtrl.InsertColumn(0, "Address");
	m_editListCtrl.SetColumnWidth(0, 100);

	// Add data columns to the list control.
	int iColumn = 1;
	CString strOffset;
	switch(m_HexEditMode)
	{
		case hemByte:
			// Configure the display for 16 bytes.
			for(int iOffset=0; iOffset<16; iOffset++)
			{
				// Insert the column.
				strOffset.Format(m_bHexadecimal ? "+%02xh" : "+%d", iOffset);
				m_editListCtrl.InsertColumn(iColumn, strOffset);

				// Set the column width.
				m_editListCtrl.SetColumnWidth(iColumn++, 40);
			}
			break;

		case hemWord:
			// Configure the display for 8 words.
			for(int iOffset=0; iOffset<8; iOffset++)
			{
				// Insert the column.
				strOffset.Format(m_bHexadecimal ? "+%02xh" : "+%d", iOffset*2);
				m_editListCtrl.InsertColumn(iColumn, strOffset);

				// Set the column width.
				m_editListCtrl.SetColumnWidth(iColumn++, 75);
			}
			break;

		case hemDword:
			// Configure the display for 4 dwords.
			for(int iOffset=0; iOffset<4; iOffset++)
			{
				// Insert the column.
				strOffset.Format(m_bHexadecimal ? "+%02xh" : "+%d", iOffset*4);
				m_editListCtrl.InsertColumn(iColumn, strOffset);

				// Set the column width.
				m_editListCtrl.SetColumnWidth(iColumn++, 100);
			}
			break;

	}

	// Add the ASCII column.
	m_editListCtrl.InsertColumn(iColumn, "ASCII");
	m_editListCtrl.SetColumnWidth(iColumn, 100);
}

void CBarPane::OnValueEdited(int iItemEdited, int iSubItemEdited, char * pszNewValue)
{
	// Get the offset of the data that was edited. The row offset is 
	// available in column 0, the item offset is determined by the column.
	CString strOffset = m_editListCtrl.GetItemText(iItemEdited, 0);
	ULONG ulOffset = strtoul(strOffset, NULL, m_bHexadecimal ? 16 : 10);
	ulOffset += m_HexEditMode * (iSubItemEdited -1);

	// Convert the new value into numeric form.
	ULONG ulNewValue = strtoul(pszNewValue, NULL, m_bHexadecimal ? 16 : 10);

	// Write the new value to the device.
	BOOL bSuccess = m_pCachedPciFunction->WriteDeviceData(m_dwBarIndex, ulOffset, &ulNewValue, m_HexEditMode, 1);
	_ASSERTE(bSuccess);

	// Read the data back to see if it 'took'.
	ULONG ulResultingValue = 0;
	bSuccess = m_pCachedPciFunction->ReadDeviceData(m_dwBarIndex, ulOffset, &ulResultingValue, m_HexEditMode, 1);
	_ASSERTE(bSuccess);

	// Update the cache with the latest data.
	m_DataCache[ulOffset] = ulResultingValue;

	// Indicate via sound if the write failed.
	if(ulNewValue != ulResultingValue)
	{
		MessageBeep(MB_ICONEXCLAMATION);
	}

	// Force the list control to update the value.
	m_editListCtrl.SetItemCountEx(m_editListCtrl.GetItemCount());
}

void CBarPane::OnSize(UINT nType, int cx, int cy)
{
	////////////////////////////////////////////////////////////////////
	//	Resize the list control width and height so that 
	//	it covers the whole area under the tab control.
	////////////////////////////////////////////////////////////////////

	// Get the current position/size of the tab control. 
	CTabCtrl * pTabCtrl = (CTabCtrl *) m_pPciFunctionView->GetDlgItem(IDC_TAB);
	CRect rcTabCtrl;
	pTabCtrl->GetWindowRect(&rcTabCtrl);

	CListCtrl * pListCtrl = (CListCtrl *) m_pPciFunctionView->GetDlgItem(IDC_BAR0_PANE_LIST_CTRL + m_dwBarIndex);
	if(pListCtrl)
	{
		// Calculate the new height of the list control.
		// This is the client area minus the height of the tab control.
		CRect rcClient;
		m_pPciFunctionView->GetClientRect(&rcClient);
		int iTabCtrlHeight = rcTabCtrl.Height();
		int iListCtrlHeight = rcClient.Height() - iTabCtrlHeight;

		// Set the new width and height of the list control.
		pListCtrl->SetWindowPos(m_pPciFunctionView, 0 , iTabCtrlHeight, cx, iListCtrlHeight, SWP_NOOWNERZORDER | SWP_NOZORDER);
	}
}

void CBarPane::ActivatePane()
{
	// We activate the window by making it highest in the Z-order.
	CWnd * pWnd = m_pPciFunctionView->GetDlgItem(IDC_BAR0_PANE_LIST_CTRL + m_dwBarIndex);
	pWnd->SetWindowPos(&CWnd::wndTop, 0 , 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	pWnd->ShowWindow(SW_SHOW);
}

void CBarPane::DeactivatePane()
{
	// We deactivate the window by hiding it.
	CWnd * pWnd = m_pPciFunctionView->GetDlgItem(IDC_BAR0_PANE_LIST_CTRL + m_dwBarIndex);
	pWnd->ShowWindow(SW_HIDE);
}

void CBarPane::ShowPciFunction(CPciFunction * pPciFunction)
{
	// Cache the CPciFunction object so we can refresh the view 
	// when switching between hex and decimal value display.
	m_pCachedPciFunction = pPciFunction;

	// Update the hex dump list control.
	m_UpdateBarEditListCtrl(m_pCachedPciFunction);
}

void CBarPane::SetHexadecimalDisplay(BOOL bHexadecimal)
{ 
	// Set the new radix to use in the list control.
	m_bHexadecimal = bHexadecimal; 

	// Update the hex dump list control.
	m_UpdateBarEditListCtrl(m_pCachedPciFunction);
}

void CBarPane::OnContextMenu(CWnd* pWnd, CPoint point)
{
	//////////////////////////////////////////////////////////
	// The user has right-clicked in the view while our 
	// pane was active so we display our context menu.
	//////////////////////////////////////////////////////////
	
	// Create the context menu.
	BCMenu menu;
	menu.LoadMenu(IDR_CONTEXTMENU_BARPANE);
	CMenu * pContextMenu = menu.GetSubMenu(0);

	// Configure the popup menu.
	pContextMenu->CheckMenuItem(IDM_HEXADECIMAL, MF_BYCOMMAND | m_bHexadecimal ? MF_CHECKED : MF_UNCHECKED);
	pContextMenu->CheckMenuItem(IDM_8BIT_DISPLAY, MF_BYCOMMAND | (m_HexEditMode == hemByte ? MF_CHECKED : MF_UNCHECKED));
	pContextMenu->CheckMenuItem(IDM_16BIT_DISPLAY, MF_BYCOMMAND | (m_HexEditMode == hemWord ? MF_CHECKED : MF_UNCHECKED));
	pContextMenu->CheckMenuItem(IDM_32BIT_DISPLAY, MF_BYCOMMAND | (m_HexEditMode == hemDword ? MF_CHECKED : MF_UNCHECKED));

	// Show the popup menu.	
	pContextMenu->TrackPopupMenu(TPM_LEFTALIGN | TPM_LEFTBUTTON, point.x, point.y, m_pPciFunctionView);
}

void CBarPane::On8bitDisplay()
{
	if(m_HexEditMode != hemByte)
	{
		// Switch display mode.
		m_HexEditMode = hemByte;

		// Update the hex dump list control.
		m_UpdateBarEditListCtrl(m_pCachedPciFunction);

		// Clear the cache since we will now display Bytes.
		m_DataCache.RemoveAll();
	};
}

void CBarPane::On16bitDisplay()
{
	if(m_HexEditMode != hemWord)
	{
		// Switch display mode.
		m_HexEditMode = hemWord;

		// Update the hex dump list control.
		m_UpdateBarEditListCtrl(m_pCachedPciFunction);

		// Clear the cache since we will now display Words.
		m_DataCache.RemoveAll();
	};
}

void CBarPane::On32bitDisplay()
{
	if(m_HexEditMode != hemDword)
	{
		// Switch display mode.
		m_HexEditMode = hemDword;

		// Update the hex dump list control.
		m_UpdateBarEditListCtrl(m_pCachedPciFunction);

		// Clear the cache since we will now display DWords.
		m_DataCache.RemoveAll();
	};
}

DWORD CBarPane::m_GetBarSize()
{
	if(!m_pCachedPciFunction)
	{
		return 0;
	}

	// Do a BAR discovery if the BAR size retrieved from the registry is 0.
	if(m_pCachedPciFunction->m_vdwBarSizes[m_dwBarIndex] == 0)
	{
		if(::MessageBox(m_pPciFunctionView->m_hWnd,
				"To determine the BAR size the PCI Function must be probed which may\n"
				"have device-dependent side-effects. Do you want to continue?", "Warning", MB_ICONWARNING | MB_YESNO) == IDYES)
			{
				return m_pCachedPciFunction->GetBarSizeFromPciDevice(m_dwBarIndex);
			}
	}

	return m_pCachedPciFunction->m_vdwBarSizes[m_dwBarIndex];
}

void CBarPane::m_UpdateBarEditListCtrl(CPciFunction * pPciFunction)
{
	// Reconfigure the display mode.
	m_ConfigureDisplayMode();

	// Get the number of items the user will 
	// be able to edit in the list control.
	DWORD dwNumItems = m_GetBarSize() / 16;  // 16 bytes per line.
	if(dwNumItems == 0)
	{
		// We at least need one row.
		dwNumItems = 1;
	}

	// Set the "virtual size" of the list control. This determines 
	// how many callbacks for data will be made by the list control.
	m_editListCtrl.SetItemCountEx(dwNumItems);
}

void CBarPane::OnGetDispInfo(LV_DISPINFO * pDispInfo)
{
	/////////////////////////////////////////////
	// We need to provide data for the list view.
	/////////////////////////////////////////////

	// If the listbox wants something else 
	// than strings we ignore the request.
	LV_ITEM * pItem = &pDispInfo->item;
	if((pItem->mask & LVIF_TEXT) == 0)
	{
		return;
	}

	// Find out the column and row of this item.
	int iRow = pItem->iItem;
	int iColumn = pItem->iSubItem;

	// If the column is 0 we should provide the address.
	CString strItem;
	if(iColumn == 0)
	{
		if(m_bHexadecimal)
			strItem.Format("%08Xh", iRow * 16);
		else
			strItem.Format("%ud", iRow * 16);
		
		lstrcpy(pItem->pszText, (const char*) strItem);
		return;
	}

	// If this is the last column we should decode the data in 
	// all columns except the first and last in ASCII format.
	int iNumColumns = m_editListCtrl.GetHeaderCtrl()->GetItemCount();
	if(iColumn == iNumColumns - 1)
	{
		// We need to iterate over all data items in this row.
		// For each data we convert the binary value into ASCII.
		// If characters are unprintable we print '.' instead.
		for(int iSubItem=1; iSubItem<iNumColumns-1; iSubItem++)
		{
			// Get the subitem string.
			CString strSubItem = m_editListCtrl.GetItemText(iRow, iSubItem);
			
			// Convert the value into numeric form.
			ULONG ulValue = strtoul(strSubItem, NULL, m_bHexadecimal ? 16 : 10);

			switch(m_HexEditMode)
			{
				case hemByte:
					if(isprint(ulValue))
						strItem.Insert(30000, __toascii(ulValue));
					else
						strItem += ".";
					break;

				case hemWord:
					for(int iCharPos=0; iCharPos<2; iCharPos++)
					{
						UCHAR ucCharValue = ulValue >> 8*iCharPos;
						if(isprint(ucCharValue))
							strItem.Insert(30000, __toascii(ucCharValue));
						else
							strItem += ".";
					}
					break;
			
				case hemDword:			
					for(int iCharPos=0; iCharPos<4; iCharPos++)
					{
						UCHAR ucCharValue = ulValue >> 8*iCharPos;
						if(isprint(ucCharValue))
							strItem.Insert(30000, __toascii(ucCharValue));
						else
							strItem += ".";
					}
					break;
			}
		}

		lstrcpy(pItem->pszText, (const char*) strItem);
		return;
	}

	// This is a data item. Calculate how many 
	// bytes we need to read from the device.
	DWORD dwItemSize = m_HexEditMode; 

	// Calculate the offset this item resides at.
	DWORD dwOffset = iRow * 16 + (iColumn - 1) * dwItemSize;

	// Do we have this data item cached?
	DWORD dwData = 0;
	BOOL bSuccess = m_DataCache.Lookup(dwOffset, dwData);
	if(!bSuccess)
	{
		// Read the data from the device.
		BOOL bSuccess = m_pCachedPciFunction->ReadDeviceData(m_dwBarIndex, dwOffset, &dwData, dwItemSize, 1);
		_ASSERTE(bSuccess);

		// Cache this data for later use.
		m_DataCache[dwOffset] = dwData;
	}

	// Format the data accordingly to the current display mode.
	switch(m_HexEditMode)
	{
		case hemByte:
			if(m_bHexadecimal)
				strItem.Format("%02Xh", dwData);
			else
				strItem.Format("%ud", dwData);
			break;
			
		case hemWord:
			if(m_bHexadecimal)
				strItem.Format("%04Xh", dwData);
			else
				strItem.Format("%ud", dwData);
			break;
			
		case hemDword:
			if(m_bHexadecimal)
				strItem.Format("%08Xh", dwData);
			else
				strItem.Format("%ud", dwData);
			break;
	}

	// Return the data and the offset.
	lstrcpy(pItem->pszText, (const char*) strItem);
}

void CBarPane::OnRefresh()
{
	// Update the hex dump list control.
	m_UpdateBarEditListCtrl(m_pCachedPciFunction);

	// Clear the data cache since we 
	// should read from the device.
	m_DataCache.RemoveAll();
}